# keg.py {{{1
# Andy Hammerlindl  2009/01/31
#
# Entry for the 2009 Game Making Deathmatch.
# A puzzle game where cannons, powderkegs, and plants lie on a grid.  The goal
# is to detonate all of the kegs without destroying the plants.

# Imports {{{1
import os
import sys
import pickle as pickle
import platform
import anydbm
import shelve
import itertools
import math
import random
import glob
from pprint import pprint
from math import sqrt, exp, log, sin, cos, tan, asin, acos, atan, atan2
from operator import add, concat

import pygame
from pygame.locals import *

# Constants {{{1
fullscreen = False

screenWidth = 1024 
screenHeight = 768
#screenWidth = 1680
#screenHeight = 1050

# Update the world atmost every 50 ms.
updateThreshold = 50

# Draw Layer - To draw all parts of the scene in the right order.  Lower
# numbers are drawn first.
BACKDROP_LAYER = -99
GAME_LAYER = 0
LABEL_LAYER = 99

# Draw Layers within the game.
TILE_LAYER = 0
FUSE_LAYER = 1
DOODAD_LAYER = 2
CANNONBALL_LAYER = 3

# Updater Layer - Update the game itself before objects handling the game.
PRIORITY_LAYER = -99
INGAME_LAYER = 0
CLICKABLES_LAYER = 1
OVERGAME_LAYER = 99 

# Events
PURGE_EVENT = USEREVENT + 1
MUSIC_EVENT = USEREVENT + 2

# Colors
fuseColor = (8,24,36)

# Util Functions {{{1

# A decorator for debugging.
def debug(fn):
    def new(*args, **kws):
        print "calling: ", fn.__name__
        pprint(args)
        pprint(kws)
        return fn(*args, **kws)
    new.__name__ = fn.__name__
    new.__doc__ = fn.__doc__
    return new

def interp(t, a, b):
    if type(a) == tuple:
        assert type(b) == tuple and len(a) == len(b)
        return tuple([interp(t, a[i], b[i]) for i in range(len(a))])
    else:
        return a + t * (b-a)

def prefix(pre, s):
    """Is pre a prefix of s?"""
    return pre == s[:len(pre)]

def argmin(sequence, fn):
    """Return the element x of sequence so that fn(x) is minimal.  Based on
    Peter Norvig's elementations:
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/389640"""
    return min((fn(x), x) for x in sequence)[1]

def mousePos():
    return pygame.mouse.get_pos()

def controlKeyPressed():
    return pygame.key.get_mods() & KMOD_CTRL

def bboxCenter(bbox):
    l,b,r,t = bbox
    return ((r+l)/2, (t+b)/2)

def dist2(loc1, loc2):
    x1,y1 = loc1; x2,y2 = loc2
    return (x1-x2)**2 + (y1-y2)**2

def dist(loc1, loc2):
    return sqrt(dist2(loc1, loc2))

def length2(vector):
    x,y = vector
    return x**2 + y**2

def length(vector):
    x,y = vector
    return sqrt(x**2 + y**2)

def unit(vector):
    x,y = vector
    if x == 0.0 and y == 0.0:
        return (0,0)
    else:
        l = length(vector)
        return (x/l, y/l)

def addVec(u, v):
    return u[0]+v[0], u[1]+v[1]
    
def dot(u, v):
    return u[0]*v[0] + u[1]*v[1]

def locationInBBox(location, bbox):
    x,y = location
    l,b,r,t = bbox
    return l <= x and x < r and b <= y and y < t

def bboxInBBox(box1, box2):
    return locationInBBox((box1[0], box1[1]), box2) and \
            locationInBBox((box1[2], box1[3]), box2)

def distToRange(x, lo, hi):
    if x < lo:
        return lo - x
    elif x > hi:
        return x - hi
    else:
        return 0
    
def distToBBox(location, bbox):
    x,y = location; l,b,r,t = bbox
    return length((distToRange(x,l,r), distToRange(y,b,t)))

def sideOfScreen(location):
    if location[0] < width/2:
        return 'left'
    else:
        return 'right'

def rectFromBBox(bbox):
    l,t,r,b=bbox
    return pygame.Rect(l,t,r-l,b-t)

def drawRectangle(box, color):
    #print "filling", rect, "with", color
    screen.fill(color, rectFromBBox(box))

def drawOutline(box, color, width=2):
    pygame.draw.rect(screen, color, rectFromBBox(box), width)

def drawLine(v1, v2, color, width=1):
    pygame.draw.line(screen, color, v1, v2, width)

def terminate():
    pygame.event.post(pygame.event.Event(QUIT))

# A list of functions to be called after pygame is initialized.
postInits = []

# Decorator to automatically add a function to postInits
def postInit(fn):
    postInits.append(fn)
    return fn

def callPostInits():
    for func in postInits:
        func()

# A list of functions to call at shutdown
onExits = []

def onExit(fn):
    onExits.append(fn)
    return fn

def callOnExits():
    for func in onExits:
        func()

# Settings {{{1
try:
    settings = shelve.open("settings", 'c')
    settingsError = False
except anydbm.error:
    # Use a non-persistant settings database if we can't open the settings.
    settings = {}
    settingsError = True

@postInit
def warnSettings():
    if settingsError:
        messageTicker.add("Could not create settings file.")
        messageTicker.add("Settings will not be saved.")

def defaultSetting(name, value):
    if name not in settings:
        settings[name] = value

defaultSetting('playMusic', True)
defaultSetting('playSound', True)
defaultSetting('lastSongPlayed', None)
defaultSetting('fullscreen', True)
defaultSetting('lastLevelPlayed', None)
defaultSetting('played', set())
defaultSetting('beaten', set())
defaultSetting('congratulated', False)

# PriorityList {{{1
def insort(a, x, key, lo=0, hi=None):
    """Insert x into the list a, using key to determine the order.  This assumes
    that a is already sorted.  x goes to the right-most position.  This is based
    on the bisect module (which doesn't support keys)."""
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo+hi)//2
        if key(x) < key(a[mid]): hi = mid
        else: lo = mid+1
    a.insert(lo, x)

class PriorityList:
    """PriorityList maintains a list of objects in an ordering prescribed by the
    level given when the object is added.  The objects are ordered by lowest
    level first.  The list can be accessed by the objs field."""
    def __init__(self):
        self.objs = []

    def add(self, obj, level):
        """Add a drawable object at a specified level."""
        insort(self.objs, (obj, level), key = lambda x: x[1])

    def remove(self, target):
        for i in range(len(self.objs)):
            obj, level = self.objs[i]
            if obj == target:
                # As this object could be in more than one priority list,
                # finalize should be idempotent.
                func =  getattr(target, 'finalize', None)
                if callable(func):
                    func()
                del self.objs[i]
                return

    def removeAll(self):
        for obj, level in self.objs[:]:
            self.remove(obj)

# Updater {{{1
class Updater(PriorityList):
    def __init__(self):
        PriorityList.__init__(self)
        self.paused = False
        self.pauseTime = None
        self.downTime = 0 # Number of ticks paused since the beginning of time.

        self.pauseLabel = None

        #self.dilation = 20.0
        self.dilation = 1

    def update(self, ticks, duration):
        """Update every object that needs to be updated."""
        if not self.paused:
            # Use a copy so objects can be removed while updating.
            for (obj, level) in self.objs[:]:
                obj.update(self.dilation * (ticks - self.downTime),
                           self.dilation * duration)

    def duration(self, t1, t2):
        """Returns the duration of an interval where times is given in pygame
        ticks.  This accounts for pausing."""
        assert t1 <= t2
        if self.paused:
            if t1 > self.pauseTime:
                return 0
            elif t2 > self.pauseTime:
                return t2 - self.pauseTime
        return t2 - t1
                
    def convert(self, ticks):
        if self.paused and ticks > self.pauseTime:
            ticks = self.pauseTime
        return self.dilation * (ticks - self.downTime)

    def get_ticks(self):
        # WARNING: This won't work properly if updaters are nested and more than
        # one of them pauses.
        return self.convert(pygame.time.get_ticks())
    
    def pause(self):
        if self.paused:
            return

        self.paused = True
        self.pauseTime = pygame.time.get_ticks()

        if self.pauseLabel == None:
            self.pauseLabel = Label('Paused', (width/2,height/2), 'center')
        else:
            self.pauseLabel.text = 'Paused'
        title.show()

    def unpause(self):
        if not self.paused:
            return

        self.downTime += pygame.time.get_ticks() - self.pauseTime
        self.pauseTime = None

        if self.pauseLabel:
            self.pauseLabel.text=''

        self.paused = False
        title.hide()

    def togglePause(self):
        if self.paused:
            self.unpause()
        else:
            self.pause()
updater = Updater()

# TargetManager {{{1
class TargetManager:
    """This translates action of the mouse to selections of object in the level.
    Each entity has to register itself with a bounding box (left, bottom, right,
    top) to the target manager.  Then, when the bounding box is clicked, the
    click method of the entity is called."""
    def __init__(self):
        self.permanent = True

        self.objs = {}
        self.focus = None
        self.selected = None
        self.pressed = False

    def add(self, obj, bbox):
        self.objs[obj] = bbox

    def remove(self, obj):
        if self.selected == obj:
            self.unselect()
        if self.focus == obj:
            self.focus = None
        if obj in self.objs:
            del self.objs[obj]

    def minimalBBoxes(self, pairs):
        if len(pairs) <= 1:
            return pairs

        minimal = []

        def isMinimal(pair):
            for m in minimal:
                if bboxInBBox(m[1], pair[1]):
                    return False
            return True
        def addMinimal(pair):
            for i in range(len(minimal)):
                if bboxInBBox(pair[1], minimal[i][1]):
                    del minimal[i]
            minimal.append(pair)

        for pair in pairs:
            if isMinimal(pair):
                addMinimal(pair)

        return minimal

    def closestObject(self, location, pairs):
        if len(pairs) == 0:
            return None
        if len(pairs) == 1:
            return pairs[0][0]

        def dist(location, pair):
            return dist2(location, bboxCenter(pair[1]))

        best = pairs[0]; bestDist = dist(location, pairs[0])
        for pair in pairs[1:]:
            distance = dist(location, pair)
            if distance < bestDist:
                best = pair; bestDist = distance
        return best[0]

    def objectAtLocation(self, location):
        candidates = [pair for pair in self.objs.items() \
                                    if locationInBBox(location, pair[1]) ]
        mins = self.minimalBBoxes(candidates)
        winner = self.closestObject(location, mins)
        return winner

    def updateMessageBar(self):
        if self.focus and hasattr(self.focus, 'desc'):
            print "message bar not yet implemented"
            #messageBar.text = self.focus.desc()
        #else:
            #messageBar.text = ''

    def focusOnLocation(self, location):
        oldFocus = self.focus
        self.focus = self.objectAtLocation(location)
        self.updateMessageBar()

        if self.focus != None and self.focus != oldFocus and \
                hasattr(self.focus, 'clickNoise'):
            mixer.play('click')

    def unselect(self):
        self.focus = None
        self.selected = None

    def click(self, location):
        assert not self.pressed
        self.pressed = True

        if self.selected:
            self.selected.click(location)
        else:
            self.focusOnLocation(location)
            if self.focus:
                self.selected = self.focus
                self.selected.click(location)

    def drag(self, location):
        if self.selected:
            drag = getattr(self.selected, 'drag', None)
            if drag:
                drag(location)

    def move(self, location):
        if self.pressed:
            self.drag(location)
        elif self.selected == None:
            self.focusOnLocation(location)

    def release(self, location):
        """Tell the selected object that the mouse is released.  If the object's
        release method returns True, it means that the object is done being
        selected.  Otherwise, it will continue to receive new clicks.
        This result is also returned, in case this TargetManager is nested in
        another TargetManager."""
        assert self.pressed
        self.pressed = False
        
        if self.selected:
            objDone = self.selected.release(location)
            if objDone:
                self.selected = None
                self.focusOnLocation(location)
            return objDone
        else:
            return True


    def key(self, unicode, key, mod):
        """Receive a key event."""
        if self.selected and hasattr(self.selected, 'key'):
            return self.selected.key(unicode, key, mod)
        else:
            return False
                
clickables = TargetManager()

# Sprite {{{1
class Sprite:
    """This is an Andy-style sprite, as opposed to the builtin Sprite module in
    pygame.  The main call is draw((x,y)) that draws the sprite at (x,y) on the
    screen.  Frames and coloring tinting is not supported, but may be added
    later."""

    def __init__(self, name, colorkey=None):
        surface = pygame.image.load(os.path.join('data', name+'.png'))
        self.surface = surface.convert_alpha()
        self.width = self.surface.get_width()
        self.height = self.surface.get_height()
        #self.surface.set_colorkey(colorkey)

    def draw(self, location):
        screen.blit(self.surface, location)

    def drawCentered(self, location):
        self.draw((location[0]-0.5*self.width, location[1]-0.5*self.height))

tileNames = [
        'tile0', 'tile1', 'backdrop']
spriteNames = [
        'barrel', 'eraser', 'cannon', 'fuse', 'plant', 'cannonball',
        'explosion', 'broken_plant', 'title']

directions = ['CW', 'CCW']
spriteNames += ['rotate'+dir for dir in directions]
compassPoints = ['E', 'N', 'W', 'S']
spriteNames += ['cannon'+str(n) for n in range(4)]

class SpriteFactory:
    """This loads all of the sprites in at the start, so they can be accessed by
    name."""
    def __init__(self):
        self.db = {}
        for name in tileNames:
            self.add(name)
        for name in spriteNames:
            self.add(name, colorkey=(0,0,0))

    def add(self, name, colorkey=None):
        self.db[name] = Sprite(name, colorkey)

    def get(self, name):
        return self.db[name]

@postInit
def initSprites():
    global sprites
    sprites = SpriteFactory()

# Scene {{{1
class Scene(PriorityList):
    """To make it easy to have special effects such as explosions, all drawing
    is managed in a central place.  The scene maintains a list of drawable
    objects (sorted by level).  To draw everything, the scene calls the draw
    method of every object in its list, which draws them in increasing order.
    
    Draw objects must support a draw() method (this will be passed timing info
    eventually), and a doneDrawing() method which returns a boolean telling the
    drawManager if it can remove the object from its list."""

    def draw(self):
        """Draw the entire scene."""
        for (obj, level) in self.objs:
            obj.draw()

    def purge(self):
        """Remove any animations or objects that no longer need to be drawn."""
        self.objs = [pair for pair in self.objs if not pair[0].doneDrawing()]

scene=Scene()


# Mixer {{{1
soundList = ['click', 'clunk', 'break', 'fuse', 'chime']
explodeList= [ 'explode'+str(i) for i in range(4) ]
reportList = [ 'report'+str(i) for i in range(3) ]
soundList += explodeList + reportList

musicList = (['music'+str(i)+'.ogg' for i in range(3) ])



class Mixer:
    def __init__(self):
        if not self.initialized():
            messageTicker.add(
                    "Cannot initialize mixer. Proceeding without sound.")
            return

        self.sounds = {}
        for name in soundList:
            self.load(name)

        self.sounds['click'].set_volume(0.15)

        random.shuffle(musicList)

        #pygame.mixer.init(44100)
        pygame.mixer.music.set_volume(0.7)
        pygame.mixer.music.set_endevent(MUSIC_EVENT)

        # If we pick a song randomly each time we start, chances are that people
        # will hear repeat songs a lot when starting up the progam.  This avoids
        # that.
        if musicList[0] == settings['lastSongPlayed']:
            self.advanceMusicList()
        self.newSong()

        if not settings['playSound']:
            messageTicker.add("Press S to enable sound effects.")
        if not settings['playMusic']:
            pygame.mixer.music.pause()
            messageTicker.add("Press M to enable music.")


    def advanceMusicList(self):
        if not self.initialized():
            return

        global musicList
        # Put the first track at the end of the list.
        musicList = musicList[1:] + musicList[0:1]

    def tagSongAsPlayed(self):
        settings['lastSongPlayed'] = self.song

    def newSong(self):
        if not self.initialized():
            return

        music = pygame.mixer.music
        self.song = musicList[0]
        music.load(os.path.join('data', self.song))
        music.play()

        # Only mark this song as the last played if it is unpaused and in fact
        # playing.
        if settings['playMusic']:
            self.tagSongAsPlayed()
        self.advanceMusicList()

    def toggleSound(self):
        settings['playSound'] = not settings['playSound']
        if not settings['playSound']:
            pygame.mixer.stop()
        messageTicker.toggleMessage(settings['playSound'], "Sound effects %s.")

    def toggleMusic(self):
        if not self.initialized():
            return

        music = pygame.mixer.music
        if settings['playMusic']:
            music.pause()

        else:
            music.unpause()

            # The program does not cycle through it songs if it starts out
            # paused.  This is ensures that a song started by unpausing is not
            # repeated in the next iteration.
            self.tagSongAsPlayed()

        settings['playMusic'] = not settings['playMusic']
        messageTicker.toggleMessage(settings['playMusic'], "Music %s.")

    def initialized(self):
        if pygame.mixer.get_init():
            return True
        else:
            return False

    def load(self, name, ext="wav"):
        if not self.initialized():
            return

        sound = pygame.mixer.Sound(os.path.join('data', name+'.'+ext))
        if not prefix('report', name) and not name in ['fuse']:
            sound.set_volume(0.3)
        self.sounds[name] = sound

    def get(self, name):
        if not self.initialized():
            return None
        return self.sounds[name]

    def play(self, name, loops=0, maxtime=0):
        if not self.initialized():
            return None
        if settings['playSound']:
            return self.sounds[name].play(loops, maxtime)
        else:
            return None

    def playRandom(self, pre, loops=0, maxtime=0):
        name = random.choice([name for name in soundList if prefix(pre, name)])
        return self.play(name, loops, maxtime)

    def stop(self, name):
        if not self.initialized():
            return None
        self.sounds[name].stop()

@postInit
def initMixer():
    global mixer
    mixer = Mixer()
    
# Game {{{1
def levelsByPrefix(prefix):
    files = glob.glob(prefix+'*.map')
    numfiles = [(int(file[len(prefix):-4]), file) for file in files]
    numfiles.sort()
    return [file for (num, file) in numfiles]

def numbersByPrefix(prefix):
    files = glob.glob(prefix+'*.map')
    numbers = [int(file[len(prefix):-4]) for file in files]
    numbers.sort()
    return numbers

tutFiles = levelsByPrefix('tut')
levelFiles = levelsByPrefix('level')
allLevelFiles = tutFiles + levelFiles

class Game:
    """Some the resources of the program, such as the menu, last for the entire
    run of the process, whereas most of the game objects are specific to a
    specific instance of the game.  This class holds its own scene, updater, etc
    with the game elements, so that when a game is ended, those resources can be
    quickly and easily disposed of."""
    def __init__(self):
        # Add a scene to the global scene.
        self.scene = Scene()
        scene.add(self.scene, GAME_LAYER)

        self.clickables = TargetManager()
        clickables.add(self.clickables, (0,0,screenWidth, screenHeight))

        self.updater = Updater()
        updater.add(self.updater, GAME_LAYER)

        self.staticMap = None
        self.dynamicMap = None

    def end(self):
        self.scene.removeAll()
        scene.remove(self)
        clickables.remove(self)

    def editMap(self, filename):
        # TODO: Don't give dimensions.
        if self.staticMap.__class__  == UserMap:
            self.editCurrentMap()
        self.staticMap.load(filename)
        self.mode = 'edit'
        showAppropriateWidgets()

    def playCurrentMap(self):
        map = self.staticMap
        if map.__class__ == StaticMap:
            self.staticMap = UserMap(map)
        else:
            assert map.__class__ == UserMap
            map.resume()

        settings['lastLevelPlayed'] = map.filename
        if map.filename in allLevelFiles:
            played = settings['played']
            played.add(map.filename)
            settings['played'] = played

        self.mode = 'play'
        showAppropriateWidgets()

    def playMap(self, filename):
        self.editMap(filename)
        self.playCurrentMap()

    def testCurrentMap(self):
        self.editCurrentMap()
        map = self.staticMap
        assert map.__class__ == StaticMap
        self.staticMap = UserMap(map)
        self.mode = 'test'
        showAppropriateWidgets()

    def testMap(self, filename):
        self.editMap(filename)
        self.testCurrentMap()

    def editCurrentMap(self):
        map = self.staticMap
        if map.__class__ == UserMap:
            map.suspend()
            map = self.staticMap = map.base
        if map.suspended:
            self.staticMap.resume()
        self.mode = 'edit'
        showAppropriateWidgets()

    def nextLevel(self):
        """Return the filename of the nextlevel."""
        thisLevel = self.staticMap.filename
        files = allLevelFiles
        if not thisLevel or thisLevel not in files:
            return None
        else:
            index = files.index(thisLevel) + 1
            return files[index] if index < len(files) else None
        
game=Game()

# Widget {{{1
class Widget(object):
    def __init__(self):
        self.permanent = True # Don't delete when the game is reset
        self.shown = False
        self.layer = LABEL_LAYER

    def show(self):
        if not self.shown:
            scene.add(self, self.layer)
            self.shown = True

    def hide(self):
        scene.remove(self)
        self.shown = False

# Label {{{1
@postInit
def initFont():
    global font
    font = pygame.font.Font(os.path.join('data', 'serif.ttf'), 20)

def justifyX(x, width, justification):
    if justification == 'left':
        return x
    elif justification in ['center', 'centre']:
        return x+width/2
    else:
        assert justification == 'right'
        return x+width

def justifyRect(rect, location, justification):
    """Modify rect representing the bounding box of a string to where the string
    would be drawn on the screen."""
    x,y=location
    rect.top=y
    if justification == 'left':
        rect.left=x
    elif justification in ['center', 'centre']:
        rect.centerx=x
    else:
        assert justification == 'right'
        rect.right=x

def stringBBox(str, location, justification):
    """The bounding box for a string at a given location."""
    w,h = font.size(str)
    rect = pygame.Rect(0,0,w,h)
    justifyRect(rect, location, justification)
    x,y,w,h = rect
    return (x,y,x+w,y+h)

def drawString(str, location, color=(255,255,255), justification='center'):
    if hasattr(str, '__iter__'):
        for s in str:
            drawString(s, location, color, justification)
            location = (location[0], location[1] + font.get_height())
    else:
        surface = font.render(str, True, color)
        rect = surface.get_rect()
        justifyRect(rect, location, justification)
        screen.blit(surface, rect)

labelColor = (100,0,0)

class Label(Widget):
    def __init__(self, text, location, justification='center'):
        Widget.__init__(self)

        self.text = text
        self.location = location
        assert justification in ['left', 'center', 'centre', 'right']
        self.justification = justification
        self.done = False
        #self.color = (255,255,255)
        self.color = labelColor

        self.resize()

    def resize(self):
        self.bbox = stringBBox(self.text, self.location, self.justification)

    def toggle(self):
        if self.shown:
            self.hide()
        else:
            self.show()

    def drawText(self, text):
        drawString(text, self.location, self.color, self.justification)

    def draw(self):
        self.drawText(self.text)

    def doneDrawing(self):
        return self.done

class WatchLabel(Label):
    def __init__(self, location, callback, filter=str, justification='center'):
        self.callback = callback
        self.filter = filter
        Label.__init__(self, filter(callback()), location, justification)

    def draw(self):
        self.drawText(self.filter(self.callback()))

#@postInit
#def initWatcher():
#    global watcher
#    watcher = WatchLabel((screenWidth/2, 10), updater.get_ticks)
#    watcher.show()
#messageBar = Label([],
#        (screenWidth/2, screenHeight-32), 'center')
# add init to show messageBar?

# Button {{{1
class ClickableWidget(Widget):
    def show(self):
        if not self.shown:
            clickables.add(self, self.bbox)
        Widget.show(self)

    def hide(self):
        clickables.remove(self)
        Widget.hide(self)

standardColorScheme = {
        'selected': ((150, 150, 150), (100, 100, 100)),
        'focus': ((255, 255, 255), (198, 198, 198)),
        'normal': ((200, 200, 200), (128, 128, 128))}
nonstandardColorScheme = {
        'selected': ((150, 150, 150), (100, 100, 100)),
        'focus': ((255, 255, 255), (198, 198, 198)),
        'normal': ((200, 200, 200), (136, 135, 125))}
beatenColorScheme = {
        'selected': ((150, 150, 150), (100, 100, 100)),
        'focus': ((255, 255, 255), (198, 198, 198)),
        'normal': ((200, 200, 200), (188, 137, 108))}
playedColorScheme = {
        'selected': ((150, 150, 150), (100, 100, 100)),
        'focus': ((255, 255, 255), (198, 198, 198)),
        'normal': ((200, 200, 200), (191, 160, 99))}

class Button(ClickableWidget):
    border = 4

    def __init__(self, text, location, onClick,
            colorSchemeCallback=None, justification='center'):
        """A text button with onClick as a callback function."""
        ClickableWidget.__init__(self)

        self.location = location
        self.clickNoise = True
        self.onClick = onClick
        self.colorSchemeCallback = colorSchemeCallback
        self.justification = justification
        self.text = text
        self.resize()

    def resize(self):
        l,t,r,b = stringBBox(self.text, self.location, self.justification)
        border = self.border

        self.bbox = (l-border, t, r+border, b + 2*border)

        if self.shown:
            clickables.add(self, self.bbox)

    def draw(self):
#        if clickables.selected == self:
#            # Clicked colors
#            fgColor = (150, 150, 150)
#            bgColor = (100, 100, 100)
#        elif clickables.focus == self:
#            # Mouse over colors
#            fgColor = (255, 255, 255)
#            bgColor = (198, 198, 198)
#        else:
#            # Normal colors
#            fgColor = (200, 200, 200)
#            bgColor = (128, 128, 128)

        state = 'selected' if clickables.selected == self else \
                'focus' if clickables.focus == self else \
                'normal'

        scheme = self.colorSchemeCallback() if self.colorSchemeCallback else \
                 standardColorScheme
        fgColor, bgColor = scheme[state]

        #drawRoundedRectangle(self.bbox, bgColor)
        drawRectangle(self.bbox, bgColor)

        x,y = self.location
        drawString(self.text, (x,y+self.border), fgColor, self.justification)

    def doneDrawing(self):
        return False

    def click(self, location):
        pass
    def release(self, location):
        self.onClick()
        return True

    def hit(self, damage):
        pass

def mom():
    print "hi mom!"

#@postInit
def initButton():
    global button
    button = Button("Say Hi", (screenWidth/2, 100), mom)

# TextBox {{{1
class TextBox(ClickableWidget):
    def __init__(self, location, capacity, text='',
                 onEnter=None, filter=None, watch=None,
                 justification='center'):
        """A text box to enter data.
        text is the initial text.
        onEnter is the callback function when text is entered.
        A filter function can be given to translate the input from a string to
        the desired format.  If filter throws a ValueError, the input is
        ignored.
        If watch is given, it is called when drawing the box to specify the
        text (if data is not being entered)."""
        ClickableWidget.__init__(self)
        self.border = 4

        self.location = location
        self.capacity = capacity
        self.text = text
        self.onEnter = onEnter
        self.filter = filter
        self.watch = watch

        self.justification = justification

        self.width = width = font.size('n'*self.capacity)[0]
        self.height = height = font.get_height()
        x,y = self.justifiedLocation(width); border = self.border
        
        self.bbox = (x - border, y,
                     x + width + border, y + height + 2*border)

        self.clickNoise = True

    def justifiedLocation(self, width):
        x,y = self.location
        j = self.justification
        if j in ['center', 'centre']:
            x -= width/2
        elif j == 'right':
            x -= width
        else:
            assert j == 'left'
        return (x,y)

    def textSize(self):
        return font.size(self.getText())

    def selected(self):
        return clickables.selected == self

    def getUnselectedText(self):
        if self.watch:
            return str(self.watch())
        else:
            return self.text

    def getText(self):
        if self.selected():
            return self.newText
        else:
            return self.getUnselectedText()

    def drawCarat(self, color):
        if pygame.time.get_ticks() / 500 % 2 == 0:
            width, height = self.textSize()
            x,y = self.justifiedLocation(width)
            border = self.border

            drawRectangle((x + width, y + border,
                           x + width + 2, y + border + height), color)

    def drawBackground(self, bgColor):
        # TODO: Get a fancy background again.
        drawRectangle(self.bbox, bgColor)

    def draw(self):
        if self.selected():
            # Clicked colors
            fgColor = (200, 200, 200)
            bgColor = (120, 100, 100)
        elif clickables.focus == self:
            # Mouse over colors
            fgColor = (230, 230, 230)
            bgColor = (190, 170, 170)
        else:
            # Normal colors
            fgColor = (200, 200, 200)
            bgColor = (150, 130, 130)

        #drawRectangle(self.bbox, bgColor)
        self.drawBackground(bgColor)

        x,y = self.location
        drawString(self.getText(), (x,y+self.border), fgColor,
                justification=self.justification)

        if self.selected():
            self.drawCarat(fgColor)

    def doneDrawing(self):
        return False

    def processText(self):
        try:
            text = self.newText
            value = self.filter(text) if self.filter else text
            self.text = str(value)
            if self.onEnter:
                self.onEnter(value)
        except ValueError:
            # Don't change the field if it doesn't pass the filter.
            pass

    def unselect(self):
        del self.newText
        clickables.unselect()
        pygame.key.set_repeat()

    def click(self, location):
        if locationInBBox(location, self.bbox):
            pygame.key.set_repeat(500, 50)
            self.newText = self.getUnselectedText()
        else:
            self.processText()
            self.unselect()

    def release(self, location):
        return False

    def key(self, unicode, key, mod):
        if key == K_RETURN:
            self.processText()
            self.unselect()
            return True
        elif key in [K_BACKSPACE, K_DELETE]:
            if len(self.newText) > 0:
                self.newText = self.newText[:-1]
            return True
        else:
            if len(self.newText) < self.capacity:
                self.newText += unicode
            return True

def AttributeBox(location, capacity, object, attr, filter=None):
    def getter():
        return getattr(object, attr)
    def setter(val):
        setattr(object, attr, val)
    return self.addTextBox(capacity,
            onEnter=setter, watch=getter, filter=filter)

def KeyBox(location, capacity, dict, key, filter=None, justification='center'):
    def getter():
        return dict().get(key)
    def setter(val):
        dict()[key] = val
    return TextBox(location, capacity,
            onEnter=setter, watch=getter, filter=filter,
            justification=justification)

#@postInit
def initTextBox():
    global boxers
    boxers = TextBox((screenWidth/2, 150), 40, 'TextBoxers',
                     onEnter=entered, filter=upper)

# Icon {{{1
selectedColor = (160,120,80)
highlightColor = (200,150,100)

dragMessages = {
        'eraser': "Click and drag on the board to erase.",
        'fuse': "Click and drag on the board to place a fuse.",
        }

class Icon(ClickableWidget):
    def __init__(self, location, sprite, group, tag, justification='left'):
        ClickableWidget.__init__(self)
        self.sprite = sprite
        self.tag = tag

        if tag in dragMessages:
            self.dragMessage = dragMessages[tag]
        else:
            self.dragMessage = 'Click on the board to place '+tag+'.'

        self.group = group
        group.add(self)

        x = justifyX(location[0], sprite.width, justification)
        y = location[1]
        self.location = (x,y)
        self.bbox = (x, y, x+sprite.width, y+sprite.height)

        self.clickNoise = True

    def draw(self):
        self.sprite.draw(self.location)
        if self.selected():
            drawOutline(self.bbox, selectedColor)
        elif clickables.focus == self:
            drawOutline(self.bbox, highlightColor)


    def select(self):
        self.group.select(self)

    def selected(self):
        return self.group.selected == self

    def click(self, location):
        self.select()
        if controlKeyPressed() and self.tag == 'cannon':
            messageTicker.add("CTRL-click on a cannon on the map to fire it.")
        return True

    def drag(self, location):
        if not locationInBBox(location, self.bbox):
            messageTicker.add(self.dragMessage)

    def release(self, location):
        return True

class IconGroup:
    def __init__(self):
        self.icons = []
        self.selected = None

    def add(self, icon):
        self.icons.append(icon)

    def select(self, icon):
        assert icon == None or icon in self.icons
        self.selected = icon

# Title {{{1
class Title(Widget):
    def __init__(self, location):
        Widget.__init__(self)
        self.location = location
        self.sprite = sprites.get('title')

    def draw(self):
        self.sprite.drawCentered(self.location)
    
    def doneDrawing(self):
        return False

@postInit
def initTitle():
    global title
    title = Title((screenWidth/2, 100))
    title.show()

# FPS {{{1
clock = pygame.time.Clock()

@postInit
def initFpsBox():
    global fpsBox
    fpsBox = Label('', (screenWidth-10, 10), justification='right')


def showFPS():
    fpsBox.text = ["FPS:%3.0f" % (clock.get_fps()),
                   "Objects:%4d" % len(scene.objs)]

def toggleFPS():
    fpsBox.toggle()


# MessageTicker {{{1

class MessageTicker(Widget):
    def __init__(self, location):
        Widget.__init__(self)
        self.spool = []
        self.location = location

    def remove(self, msg):
        self.spool = [(t, m) for (t,m) in self.spool if m != msg]

    def add(self, msg):
        # Remove any old duplicate messages.
        self.remove(msg)

        self.spool.append((updater.get_ticks(), msg))

    def toggleMessage(self, state, msg):
        """Displays on screen when a setting is toggled.  msg is a format screen
        as in "The wingbat option has been %s." """
        messages = [msg % 'disabled', msg % 'enabled']
        i = 1 if state else 0
        self.remove(messages[1-i])
        self.add(messages[i])

    def purge(self):
        now = updater.get_ticks()
        newSpool = []
        for (ticks, msg) in self.spool:
            if now < ticks + 5000:
                newSpool.append((ticks, msg))
        self.spool = newSpool

    def messages(self):
        self.purge()
        return [msg for (ticks, msg) in self.spool]

    def draw(self):
        now = updater.get_ticks()
        x,y = self.location
        for msg in reversed(self.messages()):
            drawString(msg, (x,y), color=labelColor, justification='left')
            y -= font.get_height() + 4


messageTicker = MessageTicker((50, screenHeight - 100))

@postInit
def initMessageTicker():
    messageTicker.show()

# LevelNameWatch {{{1
@postInit
def initLevelNameWatch():
    def levelName():
        map = game.staticMap
        if map:
            return map.name()
        else:
            return ''

    global levelNameWatch
    levelNameWatch = WatchLabel((70,20), levelName, justification='left')
    levelNameWatch.show()
            
# StaticMap {{{1
def isCannon(tag):
    prefix = 'cannon'
    return tag[:len(prefix)] == prefix

class StaticMap(object):
    """An editable map that hold the game data for the positions and
    orientations for all cannons, kegs, and plants."""

    fuseOffsets = {'h': (1,0), 'v': (0,1)}

    def __init__(self, rows, columns):
        """Creates a map with the given number of tiles."""
        self.tiles = []
        for i in range(2):
            self.tiles.append(sprites.get('tile'+str(i)))
        self.initRotators()

        object.__setattr__(self, 'rows', rows)
        object.__setattr__(self, 'columns', columns)

        #self.offset = (50,50)

        self.doodads = {}
        self.extras = {}.fromkeys(['cannon', 'barrel', 'plant', 'fuse'], 0)

        self.fuses = {'h': [], 'v': []}
        self.hint = 'Hint'

        w,h = self.tileSize()
        self.fuseOffset = (w/2, 3*h/4)
        self.fuseSource = None

        # Flag for when the mouse is held down and the eraser is selected.
        self.erasing = False

        self.filename = None

        self.resume()

    def name(self):
        filename = self.filename
        if not filename:
            return ''
        elif prefix('tut', filename):
            num = filename[len('tut'):-4]
            return "Tutorial "+num
        elif prefix('level', filename):
            num = filename[len('level'):-4]
            return "Level "+num
        else:
            return filename

    # Used by DynamicMap to instance a run of this map.
    # Overloaded by UserMap.
    def allDoodads(self):
        return self.doodads.iteritems()
    def allFuses(self, dir):
        return self.fuses[dir]

    def allBarrelsPlaced(self):
        """Used when checking victory conditions to see if all of the barrels
        have placed on the map.  For a map in edit mode, the point is moot, so
        return True."""
        return True

    def reset(self):
        """Remove all added elements.  For a UserMap, only the base map objects
        will remain."""
        self.doodads.clear()
        for dir in ['h', 'v']:
            del self.fuses[dir][:]

    def useString(self, tag):
        # No limits on usage in base map.
        return ''

    def suspend(self):
        game.clickables.remove(self)
        game.scene.remove(self)
        self.suspended = True

    def resume(self):
        self.resize()
        game.scene.add(self, TILE_LAYER)
        self.suspended = False

    def tileSize(self):
        t = self.tiles[0]
        return (t.width, t.height)

    def relocate(self):
        center = (screenWidth/3, screenHeight/2 - 50)
        tw, th = self.tileSize()
        width = self.columns * tw; height = self.rows * th
        self.offset = (max(center[0] - width/2, 50),
                max(center[1] - height/2, 50))

    def resize(self):
        self.relocate()
        left, top = self.locationFromTile((0,0))
        right, bottom = self.locationFromTile((self.columns, self.rows))
        self.bbox = (left, top, right, bottom)

        def inRange(tile):
            return tile[0] < self.columns and tile[1] < self.rows

        for tile in self.doodads.keys():
            if not inRange(tile):
                del self.doodads[tile]

        for dir,fuses in self.fuses.items():
            offset = self.fuseOffsets[dir]
            self.fuses[dir] = \
                    [fuse for fuse in fuses if inRange(addVec(fuse, offset))]

        #game.clickables.remove(self)
        game.clickables.add(self, self.bbox)

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)
        if name in ['rows', 'columns']:
            self.resize()

    def drawDoodad(self, tile, tag):
        sprites.get(tag).draw(self.locationFromTile(tile))

    def initRotators(self):
        w,h = self.tileSize()
        border = 5

        self.rotators = {}

        sprite = sprites.get('rotateCW')
        ww = sprite.width; hh = sprite.height
        self.rotators['CW'] = (border, h-border-hh, border+ww, h-border)
        sprite = sprites.get('rotateCCW')
        ww = sprite.width; hh = sprite.height
        self.rotators['CCW'] = (w-border-ww, h-border-hh, w-border, h-border)

    def drawRotators(self, tile):
        x,y = self.locationFromTile(tile)
        for dir, box in self.rotators.items():
            sprites.get('rotate'+dir).draw((x+box[0],y+box[1]))

    def rotatorAtLocation(self, location):
        """If there is a rotator at the tile-adjusted coordinates."""
        for dir, box in self.rotators.items():
            if locationInBBox(location, box):
                return dir
        return None

    def canRotate(self, tile):
        return tile and tile in self.doodads and \
                isCannon(self.doodads[tile]) and \
                (self.icon() == None or \
                isCannon(self.icon().tag))

    def rotateCannon(self, tile, dir):
        offset = {'CW': -1, 'CCW' : 1}
        tag = self.doodads[tile]
        assert isCannon(tag)
        n = int(tag[-1:])
        self.doodads[tile] = 'cannon'+str((n+offset[dir])%4)

    def fireCannon(self, tile, map=None, preach=True):
        if tile in self.doodads and \
                isCannon(self.doodads[tile]):
            # The DynamicMap initializer will handle the rest.
            game.dynamicMap = DynamicMap(map if map else self, tile)
            showAppropriateWidgets()
            return True
        else:
            if preach:
                messageTicker.add("CTRL key down.")
            return False

    def drawFuse(self, tile, offset):
        loc = self.locationFromTile(tile)
        drawLine(self.fuseLocationFromTile(tile),
                self.fuseLocationFromTile(addVec(tile,offset)),
                fuseColor, width=2)

    def drawFuses(self):
        for key in ['h', 'v']:
            offset = self.fuseOffsets[key]
            for tile in self.allFuses(key):
                self.drawFuse(tile, offset)

        if self.fuseSource:
            # Draw the fuse being laid by the mouse.
            mouseTile = self.tileFromLocation(mousePos())
            if self.fusePossible(mouseTile, self.fuseSource):
                color = fuseColor
            else:
                color = (180, 24, 36) # Red for impossible.
            drawLine(self.fuseLocationFromTile(self.fuseSource),
                mousePos(),
                color, width=2)

    def numFuses(self):
        return len(self.fuses['h']) + len(self.fuses['v'])

    def addDirFuse(self, tile, dir):
        if tile not in self.fuses[dir]:
            self.fuses[dir].append(tile)

    def enoughFuse(self, length):
        """Used by UserMap to check if there is enough fuse allocated to lay
        down a certain length."""
        return True

    def fusePossible(self, tile1, tile2):
        if tile1 and tile2:
            x1, y1 = tile1; x2, y2 = tile2
            return (x1 == x2 or y1 == y2) and \
                    self.enoughFuse(dist(tile1, tile2))
        else:
            return False

    def addFuse(self, tile1, tile2, testing=False):
        if tile1 and tile2:
            x1, y1 = tile1; x2, y2 = tile2
            if y1 == y2:
                if x1 == x2:
                    return
                elif x2 < x1:
                    x1, x2 = x2, x1
                if self.enoughFuse(x2-x1):
                    for x in range(x1, x2):
                        self.addDirFuse((x,y1), 'h')
                else:
                    messageTicker.add("You do not have enough fuse.")
            elif x1 == x2:
                if y2 < y1:
                    y1, y2 = y2, y1
                if self.enoughFuse(y2-y1):
                    for y in range(y1, y2):
                        self.addDirFuse((x1,y), 'v')
                else:
                    messageTicker.add("You do not have enough fuse.")
            else:
                messageTicker.add("Fuses cannot be laid diagonally.")

    def numDoodads(self, tag):
        if tag == 'cannon':
            return len(filter(isCannon, self.doodads.values()))
        else:
            return self.doodads.values().count(tag)

    def addDoodad(self, tile, tag):
        self.doodads[tile] = tag

    def drawMouseOver(self):
        tile = self.tileFromLocation(mousePos())
        if self.canRotate(tile):
            self.drawRotators(tile)

    def drawFloor(self):
        for i in range(self.columns):
            for j in range(self.rows):
                self.tiles[(i+j)%2].draw(self.locationFromTile((i,j)))

    def drawDoodads(self):
        for tile, tag in self.allDoodads():
            self.drawDoodad(tile, tag)

    def draw(self):
        self.drawFloor()
        self.drawFuses()

        self.drawDoodads()

        self.drawMouseOver()

        outlineColor = (210,130,40) if controlKeyPressed() else (100,0,0)
        drawOutline(self.bbox, outlineColor)


    savekeys = ['rows', 'columns', 'doodads', 'fuses', 'hint', 'extras']

    def save(self, filename):
        #db = shelve.open(filename, 'n')
        db = {}
        for key in self.savekeys + ['filetype', 'version']:
            db[key] = getattr(self, key)

        f = file(filename, 'w')
        pickle.dump(db, f)
        f.close()

        #db.close()
        self.filename = filename

    filetype = "Powderkeg map"; version = (1,0,0)

    def load(self, filename):
        try:
            #db = shelve.open(filename, 'r')
            f = file(filename, 'r')
            db = pickle.load(f)
            f.close()
        #except anydbm.error:
        except IOError:
            messageTicker.add("Cannot open map '"+filename+"'.")
            return
        except KeyError:
            messageTicker.add("File '"+filename+"' is not a Powderkeg map.")
            return

        if 'filetype' not in db or db['filetype'] != self.filetype:
            messageTicker.add("File '"+filename+"' is not a Powderkeg map.")
            return
        if 'version' not in db or db['version'] != self.version:
            messageTicker.add(
                    "Map '"+filename+"' is of an incompatible version.")
            return

        for key in self.savekeys:
            if key in db:
                setattr(self, key, db[key])
        self.filename = filename

    def locationFromTile(self, tile):
        w,h = self.tileSize()
        return (w*tile[0]+self.offset[0], h*tile[1]+self.offset[1])

    def fuseLocationFromTile(self, tile):
        return addVec(self.locationFromTile(tile), self.fuseOffset)

    def tileFromLocation(self, location):
        w,h = self.tileSize()
        i = (location[0]-self.offset[0]) / w
        j = (location[1]-self.offset[1]) / h

        if i >= 0 and i < self.columns and j >= 0 and j < self.rows:
            return i,j
        else:
            return None

    def icon(self):
        """Returns the selected icon in the side bar."""
        return getIcon()

    def offsetFromLocation(self, location):
        origin = self.locationFromTile(self.tileFromLocation(location))
        return (location[0] - origin[0], location[1] - origin[1])

    def cannonMessage(self):
        # Don't harass the user in edit mode.
        pass

    def clickRotate(self, tile, location):
        offset = self.offsetFromLocation(location)
        dir = self.rotatorAtLocation(offset)
        if dir:
            self.rotateCannon(tile, dir)
        else:
            self.cannonMessage()

    def click(self, location):
        tile = self.tileFromLocation(location)

        # For now, fire a cannon by CTRL-clicking.
        if controlKeyPressed():
            return self.fireCannon(tile)

        # Let the rotate buttons on the cannons see the click first.
        if self.canRotate(tile):
            self.clickRotate(tile,location)
            return True

        icon = self.icon()

        if self.cannonAtTile(tile) and icon == None or icon == 'cannon':
            self.cannonMessage()
            return True

        if tile and icon:
            method = getattr(self, 'click_'+icon.tag, None)
            if method:
                return method(tile, location)
            else:
                self.addDoodad(tile=tile, tag=icon.tag)

        return True

    def cannonAtTile(self, tile):
        return tile in self.doodads and isCannon(self.doodads[tile])

    def click_cannon(self, tile, location):
        # Translate a 'cannon' tag to a 'cannon2' tag.
        if self.cannonAtTile(tile):
            self.cannonMessage()
        self.addDoodad(tile=tile, tag='cannon2')
        return True

    def click_eraser(self, tile, location):
        self.erasing = True
        self.drag_eraser(location)
        return False

    def click_fuse(self, tile, location):
        self.fuseSource = tile
        return True

    def drag(self, location):
        if self.erasing:
            self.drag_eraser(location)

    def drag_eraser(self, location):
        tile = self.tileFromLocation(location)
        if not tile:
            return

        # Delete any doodad in the tile.
        if tile in self.doodads:
            del self.doodads[tile]

        # Delete any fuse in the tile.
        # TODO: Use fuseOffsets.
        hf = self.fuses['h']; vf = self.fuses['v']
        tileW = (tile[0]-1, tile[1]); tileN = (tile[0], tile[1]-1)
        if tile in hf:
            hf.remove(tile)
        if tileW in hf:
            hf.remove(tileW)
        if tile in vf:
            vf.remove(tile)
        if tileN in vf:
            vf.remove(tileN)

    def release(self, location):
        if self.fuseSource:
            fuseEnd = self.tileFromLocation(location)
            self.addFuse(self.fuseSource, fuseEnd)
            if self.fuseSource == fuseEnd and self.cannonAtTile(fuseEnd):
                self.cannonMessage()
        self.fuseSource = None
        self.erasing = False
        return True

@postInit
def initMap():
    # The Game class cannot create a map directly, as Game is defined before
    # StaticMap.
    game.staticMap = StaticMap(7,7)

# UserMap {{{1
class UserMap(StaticMap):
    """A static map in which only a fixed number of cannon, barrels, and fuses
    can be added to a base map."""
    def __init__(self, base):
        StaticMap.__init__(self, base.rows, base.columns)
        self.base = base
        #self.offset = base.offset
        self.hint = base.hint
        self.filename = base.filename

        base.suspend()

    def destroy(self):
        self.suspend()
        self.base.resume()

    def allDoodads(self):
        return itertools.chain(StaticMap.allDoodads(self),
                self.base.allDoodads())

    def allFuses(self, dir):
        return itertools.chain(StaticMap.allFuses(self, dir),
                self.base.allFuses(dir))

    def maxedOutFuses(self):
        """If we have reached our limit."""
        assert self.numFuses() <= self.base.extras['fuse']
        return self.numFuses() == self.base.extras['fuse']

    def enoughFuse(self, length):
        assert self.numFuses() <= self.base.extras['fuse']
        return self.numFuses()+length <= self.base.extras['fuse']


    def addDirFuse(self, tile, dir):
        if tile not in self.base.fuses[dir]:
            if self.maxedOutFuses():
                messageTicker.add("You are out of fuses.")
                # Add sound?
            else:
                StaticMap.addDirFuse(self, tile, dir)

    def maxedOutDoodad(self, tag):
        if prefix('cannon', tag):
            tag = 'cannon'
        assert self.numDoodads(tag) <= self.base.extras[tag]
        return self.numDoodads(tag) == self.base.extras[tag]

    def allBarrelsPlaced(self):
        return self.maxedOutDoodad('barrel')

    def addDoodad(self, tile, tag):
        if tile not in self.base.doodads:
            if self.maxedOutDoodad(tag):
                if isCannon(tag):
                    messageTicker.add("You are out of cannons.")
                else:
                    messageTicker.add("You are out of "+tag+"s.")
            else:
                StaticMap.addDoodad(self, tile, tag)

    def useString(self, tag):
        used = self.numFuses() if tag == 'fuse' else self.numDoodads(tag)
        max = self.base.extras[tag]
        return str(max-used)+'/'+str(max)

    def fireCannon(self, tile, map=None, preach=True):
        assert map == None
        return StaticMap.fireCannon(self, tile, preach=False) or \
                self.base.fireCannon(tile, map=self, preach=preach)

    def cannonAtTile(self, tile):
        return StaticMap.cannonAtTile(self, tile) or \
                self.base.cannonAtTile(tile)

    def cannonMessage(self):
        messageTicker.add("CTRL-click on the cannon to fire.")

# DynamicMap {{{1
class DynamicMap:
    def __init__(self, staticMap, fireTile):
        """Simulates a run of the map, first firing the cannon at fireTile."""
        self.staticMap = staticMap

        staticMap.suspend()

        game.scene.add(self, TILE_LAYER)
        game.updater.add(self, INGAME_LAYER)

        self.startTicks = game.updater.get_ticks()

        self.scene = Scene()

        self.initDoodads()
        self.cannonballs = []
        self.initFuses()

        # Time in the dynamic map is represented as a tuple (frame, fraction).
        # A frame corresponds to the time a cannon ball take to move one tile,
        # which is the same tile a fuse takes to burn one tile (should we make
        # this 2:1 or 3:1?).
        # The fraction is a floating point number between 0.0 and 1.0 used to
        # for animation purposes to go from one frame to the next.
        self.frame = 0

        # The frame that activity ended.
        self.endFrame = None

        # If the dynamic game is being displayed in its final state, waiting for
        # the user to click on the Next Level button.
        self.frozen = False

        # The amount of time in ticks that a frame lasts.
        self.frameDuration = 1000

        # The number of ticks that have elapsed for the current frame.
        self.frameTicks = 0

        # This holds a reference to the persistant fuse sound when it is on.
        self.fuseSound = None

        # TODO: fire first cannon
        cannon = self.doodads[fireTile].fire()

    def dimensions(self):
        return self.staticMap.columns, self.staticMap.rows

    def initDoodads(self):
        classOfTag = {'plant': Plant, 'barrel': Barrel}
        for i in range(4):
            classOfTag['cannon'+str(i)] = Cannon

        self.doodads = {}
        for tile, tag in self.staticMap.allDoodads():
            cls = classOfTag[tag]
            self.doodads[tile] = cls(self, tag, tile)

    def initFuses(self):
        self.fuses = {}
        for key in ['h', 'v']:
            self.fuses[key] = {}
            offset = StaticMap.fuseOffsets[key]
            for tile in self.staticMap.allFuses(key):
                self.fuses[key][tile] = Fuse(self, tile, offset)

    def addCannonball(self, ball):
        self.cannonballs.append(ball)

    def iterfuses(self):
        """An iterator over all fuses."""
        return itertools.chain(*[fuses.itervalues() for fuses in \
            self.fuses.itervalues()])

    def objects(self):
        """An iterator over all of the objects in the map."""
        return itertools.chain(self.doodads.itervalues(),
                self.cannonballs, self.iterfuses())

    def targetsAtTile(self, tile):
        if tile in self.doodads:
            yield self.doodads[tile]
        for key, fuses in self.fuses.items():
            offset = StaticMap.fuseOffsets[key]
            if tile in fuses:
                yield fuses[tile]
            end = (tile[0] - offset[0], tile[1] - offset[1])
            if end in fuses:
                yield fuses[end]

    def draw(self):
        self.staticMap.drawFloor()
        self.scene.draw()

    def done(self):
        return self.endFrame != None and self.frame >= self.endFrame + 2

    def freeze(self):
        self.frozen = True

    def active(self):
        for obj in self.objects():
            if obj.active():
                return True
        #return ticks - self.startTicks < 5000

    def trauma(self):
        """Returns true if any plants have been broken."""
        for obj in self.objects():
            if obj.__class__ == Plant and obj.broken:
                return True
        return False

    def exploded(self):
        """Return true if all barrels have been broken."""
        if not self.staticMap.allBarrelsPlaced():
            return False
        for obj in self.objects():
            if obj.__class__ == Barrel and not obj.exploded():
                return False
        return True

    def state(self):
        """Return 'success', 'failure', or 'trauma' based on what happened
        during the run.  The state is used by Game to figure out which dialog
        box to display."""
        return 'trauma' if self.trauma() else \
                'success' if self.exploded() else 'failure'

    def shutdown(self):
        self.scene.removeAll()
        self.doodads.clear()

        if self.fuseSound:
            self.fuseSound.stop()

        game.scene.remove(self)
        game.clickables.remove(self)
        game.updater.remove(self)
        game.dynamicMap = None

    def advance(self):
        assert self.frameTicks >= self.frameDuration
        self.frame += 1
        self.frameTicks -= self.frameDuration

        if self.endFrame == None:
            for obj in self.objects():
                obj.advance()

            # Mark the animation as over if nothing is still running.
            if not self.active():
                self.endFrame = self.frame


    def action(self, field, tile):
        result = False
        for obj in self.targetsAtTile(tile):
            method = getattr(obj, field)
            if method and method(tile):
                result = True
        return result

    def hit(self, tile):
        return self.action('hit', tile)
    def spark(self, tile):
        return self.action('spark', tile)
    def shock(self, tile):
        return self.action('shock', tile)

    def fusesBurning(self):
        """Return True if any fuses are active."""
        for obj in self.objects():
            if obj.__class__ == Fuse and obj.active():
                return True
        return False

    def adjustFuseSound(self):
        if self.fusesBurning():
            if not self.fuseSound:
                self.fuseSound = mixer.play('fuse')
        else:
            if self.fuseSound:
                self.fuseSound.stop()
                self.fuseSound = None

    def update(self, ticks, duration):
        if self.frozen:
            return
        elif self.done():
            postRun()
        else:
            self.frameTicks += duration
            if self.frameTicks >= self.frameDuration:
                self.advance()

            self.adjustFuseSound()

    def time(self):
        return (self.frame, float(self.frameTicks) / float(self.frameDuration))

    def locationFromTile(self, tile):
        return self.staticMap.locationFromTile(tile)

    def fuseLocationFromTile(self, tile):
        return self.staticMap.fuseLocationFromTile(tile)

    def tileSize(self):
        return self.staticMap.tileSize()

# Doodad {{{1
class Doodad:
    """Base class for doodads (cannon, barrels, and plants) in the dynamic
    map."""
    def __init__(self, map, tag, tile):
        self.map = map
        self.tag = tag
        self.tile = tile

        self.sprite = sprites.get(tag)

        map.scene.add(self, DOODAD_LAYER)

    def location(self):
        return self.map.locationFromTile(self.tile)

    def draw(self):
        self.sprite.draw(self.location())

    def active(self):
        return False

    def advance(self):
        """Called each time a new frame starts."""
        pass

    def spark(self, tile):
        """This is called when the spark from a lit fuse reaches this doodad."""
        return False
    def shock(self, tile):
        """This is called when a barrel explodes adjacent to this doodad."""
        return False
    def hit(self, tile):
        """This is called when a cannonball hits this doodad."""
        #print "hit", self.tag, "at", self.tile
        return False

class Cannon(Doodad):
    def __init__(self, map, tag, tile):
        Doodad.__init__(self, map, tag, tile)
        self.fired = False

    def fire(self):
        if not self.fired:
            self.fired = True
            ball = Cannonball(self.map, self.tile, velocityFromTag(self.tag))
            self.map.addCannonball(ball)
            mixer.playRandom('report')

    def spark(self, tile):
        self.fire()
        return True

    def hit(self, tile):
        # TODO: Have DynamicMap compile sounds together on advance.
        mixer.play('clunk')
        return True

class Barrel(Doodad):
    def __init__(self, map, tag, tile):
        Doodad.__init__(self, map, tag, tile)
        self.explodeFrame = None


    def draw(self):
        frame, frac = self.map.time()
        if self.explodeFrame == None:
            Doodad.draw(self)
        elif self.explodeFrame == frame and frac < 0.4:
            if frac < 0.2:
                Doodad.draw(self)
            sprites.get('explosion').draw(self.location())
        else:
            #Draw rubble.
            pass

    def explode(self):
        if not self.exploded():
            self.explodeFrame = self.map.frame
            self.map.spark(self.tile)
            for offset in velocities:
                self.map.shock(addVec(self.tile, offset))
            mixer.playRandom('explode')
            return True
        else:
            # If two cannonball contact a barrel at the same time, they both
            # must be stopped.
            return self.explodeFrame == self.map.frame

    def exploded(self):
        return self.explodeFrame != None

    def spark(self, tile):
        return self.explode()

    def shock(self, tile):
        return self.explode()

    def hit(self, tile):
        return self.explode()

class Plant(Doodad):
    def __init__(self, map, tag, tile):
        Doodad.__init__(self, map, tag, tile)
        self.broken = False

    def fracture(self):
        if not self.broken:
            self.broken = True
            self.sprite = sprites.get('broken_plant')
            mixer.play('break')

    def hit(self, tile):
        self.fracture()
        return True

    def shock(self, tile):
        self.fracture()
        return True

# Cannonball {{{1
velocities = [(1,0),(0,-1),(-1,0),(0,1)]
def velocityFromTag(tag):
    """Returns the correct velocity for a cannonball fired from a cannon with
    the given tag."""
    assert isCannon(tag)
    return velocities[int(tag[-1:])]

class Cannonball:
    def __init__(self, map, startTile, velocity):
        self.map = map
        self.startFrame = map.frame
        self.startTile = startTile
        self.velocity = velocity
        self.flying = True

        self.sprite = sprites.get('cannonball')

        map.scene.add(self, CANNONBALL_LAYER)

        # Collide immediately with anything right in front of the cannon.
        self.advance()

    def active(self):
        return self.flying

    def positionAtFrame(self, frame):
        x,y = self.startTile
        vx,vy = self.velocity
        offset = frame - self.startFrame + 1
        return (x+offset*vx, y+offset*vy)

    def positionAtTime(self, t):
        x,y = self.startTile
        vx,vy = self.velocity

        t += 0.5 - self.startFrame
        return (x+t*vx, y+t*vy)

    def draw(self):
        if self.active():
            frame, frac = self.map.time()
            t = frame + frac
            x,y = self.positionAtTime(t)
            self.sprite.drawCentered(self.map.locationFromTile((0.5+x,0.5+y)))

    def advance(self):
        if self.active():
            pos = self.positionAtFrame(self.map.frame)
            w, h = self.map.dimensions()
            if self.map.hit(pos) or not locationInBBox(pos, (0,0,w,h)):
                # Stop once a target is hit.
                self.flying = False

# Fuse {{{1
class Fuse:
    def __init__(self, map, tile, offset):
        """A fuse starting at tile and going to tile+offset."""
        self.map = map
        self.tile = tile
        self.offset = offset

        # The frames in which the start and end of the fuse are lit.
        self.litFrames = [None, None]

        map.scene.add(self, FUSE_LAYER)

    def tileAt(self, t):
        x,y = self.tile; dx, dy = self.offset
        return (x+t*dx,y+t*dy)

    def location(self, t):
        return self.map.fuseLocationFromTile(self.tileAt(t))

    def drawInterval(self, t0, t1):
        if t0 < t1:
            drawLine(self.location(t0), self.location(t1),
                fuseColor, width=2)

    def burnPosition(self, frame):
        """Return an number from 0.0 to 1.0 indicating how much of a fuse
        lit at the given frame has burned."""
        now, frac = self.map.time()
        return 0.0  if frame == None else \
               frac if frame == now  else \
               1.0;

    sparkOffsets = [
            (-2,0), (-2,-1),
            (-1,0), (-1,-1), (-1,-2), (-1,-3),
            (0,0), (0,-1), (0,-2), (0,-3),
            (1,0), (1,-1), (1,-2), (1,-3),
            (2,0), (2,-1)]
    sparkColors = [
            (255, 255, 255),
            (255, 200, 200),
            (255, 255, 100)]

    def drawSpark(self, location):
        rnd = random.randint
        #offset = (rnd(-2,2), rnd(-3,0))
        offset = random.choice(self.sparkOffsets)
        color = random.choice(self.sparkColors)
        drawLine(location, addVec(location, offset), color)

    def burnInterval(self):
        t = [self.burnPosition(self.litFrames[i]) for i in [0,1]]
        t[1] = 1.0 - t[1]
        return t

    def draw(self):
        t = self.burnInterval()
        self.drawInterval(t[0], t[1])
        if t[0] <= t[1]:
            for i in [0,1]:
                if self.activeEnd(i):
                    self.drawSpark(self.location(t[i]))

    def activeEnd(self, i):
        assert i in [0,1]
        lit = self.litFrames; frame = self.map.frame
        return lit[i] == frame and (lit[1-i] == None or lit[i-1] == frame)

    def active(self):
        if self.activeEnd(0) or self.activeEnd(1):
            t = self.burnInterval()
            return t[0] <= t[1]

    def advance(self):
        """Called each time a new frame starts."""
        lit = self.litFrames
        frame = self.map.frame
        for i in [0,1]:
            if lit[i] != None and lit[1-i] == None and lit[i]+1 == frame:
                self.map.spark(self.tileAt(1-i))

    def light(self, tile):
        for i in [0,1]:
            if tile == self.tileAt(i) and self.litFrames[i] == None:
                self.litFrames[i] = self.map.frame
        return self.active()

    def spark(self, tile):
        return self.light(tile)

    def shock(self, tile):
        return False

    def hit(self, tile):
        return False

# SideBar {{{1
class SideBar:
    """A menu/toolbar on the right side of the screen that contains a list of
    labels, buttons, icons, etc. for editing the map."""
    def __init__(self, height):
        self.widgets = []
        self.iconGroup = IconGroup()

        self.width = width = 200
        border = 10
        self.location = (screenWidth-width-2*border, height)

        # The height of the next item to add.
        self.nextHeight = self.location[1]

        # The spacing between items.
        self.margin = 5

    def show(self):
        for widget in self.widgets:
            widget.show()

        global currentSideBar
        currentSideBar = self

    def hide(self):
        for widget in self.widgets:
            widget.hide()

    def nextLocation(self, justification):
        x = self.location[0]
        if justification == 'left':
            pass
        elif justification in ['center', 'centre']:
            x += self.width/2
        else:
            assert justification == 'right'
            x += self.width
        y = self.nextHeight

        return (x, y)

    def addWidget(self, widget):
        self.widgets.append(widget)
        self.nextHeight = max(self.nextHeight, widget.bbox[3] + self.margin)

        widget.hide()

        return widget

    def addLabel(self, text, justification='left'):
        label = Label(text,
                      self.nextLocation(justification),
                      justification)
        return self.addWidget(label)

    def addButton(self, text, onClick, justification='center'):
        button = Button(text,
                        self.nextLocation(justification),
                        onClick)
        return self.addWidget(button)

    def addTextBox(self, capacity, text='',
            onEnter=None, filter=None, watch=None,
            justification='center'):
        textBox = TextBox(self.nextLocation(justification),
                          capacity,
                          text=text,
                          onEnter=onEnter,
                          filter=filter,
                          watch=watch)
        return self.addWidget(textBox)

    def addWatchBox(self, capacity, object, field, filter=None,
            justification='center'):
        def getter():
            return getattr(object, field)
        def setter(val):
            setattr(object, field, val)
        return self.addTextBox(capacity,
                onEnter=setter, watch=getter, filter=filter,
                justification=justification)

    def addIcon(self, sprite, tag, justification='left'):
        icon = Icon(self.nextLocation(justification),
                    sprite,
                    self.iconGroup,
                    tag,
                    justification)
        return self.addWidget(icon)

    def addIconAndKeyBox(self, sprite, tag,
            capacity, dict, filter=None):
        # This is a bit of a hack.
        icon = self.addIcon(sprite, tag, 'left')

        # Position the text box just to the right of the icon.
        x = icon.bbox[2] + 50
        y = (icon.bbox[1] + icon.bbox[3] - font.get_height())/2

        keyBox = KeyBox((x,y), 2, dict, tag, filter=filter,
                justification='center')
        self.addWidget(keyBox)
        return icon

    def addIconAndWatchLabel(self, sprite, tag,
            callback):
        # This is also bit of a hack.
        icon = self.addIcon(sprite, tag, 'left')

        # Position the label just to the right of the icon.
        x = icon.bbox[2] + 50
        y = (icon.bbox[1] + icon.bbox[3] - font.get_height())/2

        # Give the tag to the callback
        def fullCallback():
            return callback(tag)
        label = WatchLabel((x,y), callback = fullCallback)
        self.addWidget(label)
        return icon

    def icon(self):
        return self.iconGroup.selected

currentSideBar = None

def getIcon():
    if currentSideBar:
        return currentSideBar.icon()
    else:
        return None

# EditBar {{{1

# UI objects that should be visible only when editing a map.
editWidgets = []
def hideWidgets(widgets):
    for widget in widgets:
        widget.hide()
def showWidgets(widgets):
    for widget in widgets:
        widget.show()

def gotoMainMenu():
    if game.staticMap:
        game.staticMap.suspend()
    showAppropriateWidgets()

@postInit
def initEditBar():
    global editBar
    editBar = SideBar(50)

    editBar.addLabel("Rows:")
    def getRow():
        return game.staticMap.rows
    def setRow(x):
        game.staticMap.rows = x
    editBar.addTextBox(2, '5', onEnter = setRow, filter = int, watch = getRow)
    #editBar.addWatchBox(4, game.staticMap, 'rows', filter = int)

    editBar.addLabel("Columns:")
    def getColumn():
        return game.staticMap.columns
    def setColumn(x):
        game.staticMap.columns = x
    editBar.addTextBox(2, '5', onEnter = setColumn, filter = int,
            watch = getColumn)

    for tag in ['barrel', 'cannon', 'fuse',]:
        editBar.addIconAndKeyBox(sprites.get(tag), tag,
                capacity=2, dict=lambda: game.staticMap.extras, filter=int)
    for tag in ['plant', 'eraser']:
        editBar.addIcon(sprites.get(tag), tag)

    def testMap():
        game.testCurrentMap()
    editBar.addButton("Test", testMap)

    editBar.addLabel("Filename:")
    filenameBox = editBar.addTextBox(20, 'mylevel.map')
    def saveMap():
        game.staticMap.save(filenameBox.text)
    editBar.addButton("Save", saveMap)
    def loadMap():
        game.staticMap.load(filenameBox.text)
    editBar.addButton("Load", loadMap)
    editBar.addButton("Menu", gotoMainMenu)

    editWidgets.append(editBar)

# UserBar {{{1

# UI objects visible only when playing a file in user mode.
userWidgets = []

@postInit
def initUserBar():
    global userBar
    userBar = SideBar(100)

    for tag in ['barrel', 'cannon', 'fuse']:
        def callback(tag):
            return game.staticMap.useString(tag)
        userBar.addIconAndWatchLabel(sprites.get(tag), tag, callback)
    for tag in ['eraser']:
        userBar.addIcon(sprites.get(tag), tag)

    def resetMap():
        game.staticMap.reset()
    userBar.addButton("Reset", resetMap)

    def endTest():
        game.editCurrentMap()
    userBar.addButton("End Test", endTest)
    userBar.addButton("Menu", gotoMainMenu)

    userWidgets.append(userBar)


# PlayBar {{{1

# UI objects visible only when playing a file in play mode.
playWidgets = []

@postInit
def initPlayBar():
    global playBar
    playBar = SideBar(100)

    for tag in ['barrel', 'cannon', 'fuse']:
        def callback(tag):
            return game.staticMap.useString(tag)
        playBar.addIconAndWatchLabel(sprites.get(tag), tag, callback)
    for tag in ['eraser']:
        playBar.addIcon(sprites.get(tag), tag)

    def resetMap():
        game.staticMap.reset()
    playBar.addButton("Reset", resetMap)
    playBar.addButton("Menu", gotoMainMenu)

#    def nextLevelCallback():
#        game.playMap(game.nextLevel())
#    playBar.addButton("Next", nextLevelCallback)

    playWidgets.append(playBar)


# HintBox and HintLabel {{{1
@postInit
def initHintBoxAndLabel():
    global hintBox, hintLabel
    def onEnter(hint):
        game.staticMap.hint = hint
    def watch():
        return game.staticMap.hint
    location = (50, screenHeight-50)
    hintBox = TextBox(location, 100, text="Give us a hint.",
            onEnter=onEnter, watch=watch,
            justification='left')
    editWidgets.append(hintBox)
    hintLabel = WatchLabel(location, watch, 
            justification='left')
    userWidgets.append(hintLabel)
    playWidgets.append(hintLabel)

# victoryWidgets {{{1
victoryWidgets = []

def replayLevel():
    game.dynamicMap.shutdown()
    game.staticMap.resume()
    showAppropriateWidgets()

@postInit
def initVictoryWidgets():
    global victoryLabel, replayButton, nextLevelButton
    victoryLabel = Label("Level Completed", (50, screenHeight-80),
            justification='left')
    victoryWidgets.append(victoryLabel)

    replayButton = Button("Revisit",
            (victoryLabel.bbox[2] + 20, screenHeight-80-Button.border),
            onClick = replayLevel,
            justification='left')
    victoryWidgets.append(replayButton)

    def nextLevelCallback():
        game.dynamicMap.shutdown()
        game.playMap(game.nextLevel())

    nextLevelButton = Button("Next Level",
            (replayButton.bbox[2] + 20, screenHeight-80-Button.border),
            onClick = nextLevelCallback,
            justification='left')
    victoryWidgets.append(nextLevelButton)

def showAppropriateWidgets():
    for widgets in [editWidgets, userWidgets, playWidgets, victoryWidgets]:
        hideWidgets(widgets)
    title.hide()
    if currentMenu == mainMenu:
        mainMenu.hide()

    if not game.dynamicMap:
        if game.staticMap.suspended:
            title.show()
            mainMenu.show()
            if not settings['lastLevelPlayed']:
                resumeButton.hide()
        elif game.staticMap.__class__ == StaticMap:
            showWidgets(editWidgets)
        else:
            assert game.staticMap.__class__ == UserMap
            assert game.mode in ['test', 'play']
            showWidgets(playWidgets if game.mode == 'play' else userWidgets)

def postRun():
    state = game.dynamicMap.state()

    if state == 'trauma':
        messageTicker.add("Don't shoot the plants!")
    elif state == 'failure':
        if game.staticMap.allBarrelsPlaced():
            messageTicker.add("Explode all of the barrels.")
        else:
            messageTicker.add("Place and explode all barrels.")

    if state == 'success':
        mixer.play('chime')

        if game.mode == 'play':
            showWidgets(victoryWidgets)
            game.dynamicMap.freeze()
            if not game.nextLevel():
                nextLevelButton.hide()

            # Update the color of the corresponding button in the main menu.
            map = game.staticMap
            if map.filename in allLevelFiles:
                beaten = settings['beaten']
                beaten.add(map.filename)
                settings['beaten'] = beaten

                if set(levelFiles) <= beaten and not settings['congratulated']:
                    replayLevel()
                    gotoMainMenu()
                    victoryMenu.show()
                    settings['congratulated'] = True
        else:
            messageTicker.add("Level completed.")
            replayLevel()
    else:
        replayLevel()

# Menu {{{1
def menuLocation(row, justification='center'):
    x = screenWidth/2
    j = justification
    if j == 'left':
        x -= 100
    elif j == 'right':
        x += 100
    else:
        assert j in ['center', 'centre']

    y = 200 + 40 * row

    return (x,y)

currentMenu = None

class Menu:
    """Bundles widgets together into a menu that can be shown and hidden en
    masse."""
    def __init__(self, parent=None):
        self.parent = parent
        self.widgets = []

        self.row = 0
        self.lastWidgetSize = 'small'

    def show(self):
        global currentMenu
        if currentMenu:
            currentMenu.hide()
        currentMenu = self

        for widget in self.widgets:
            widget.show()

    def hide(self):
        for widget in self.widgets:
            widget.hide()

        global currentMenu
        assert currentMenu == self
        currentMenu = None

    def close(self):
        self.hide()
        if self.parent:
            self.parent.show()

    def extraRows(self, size):
        assert size in ['small', 'big']
        if self.lastWidgetSize == 'big' and size == 'big':
            return 1
        else:
            return 0

    def nextLocation(self, size, justification='center'):
        row = self.row + self.extraRows(size)
        return menuLocation(row, justification)

    def addWidgetWithoutSize(self, widget):
        self.widgets.append(widget)
        widget.hide()
        return widget

    def addSpacing(self, size):
        assert size in ['small', 'big']
        self.row += 1 + self.extraRows(size)
        self.lastWidgetSize = size


    def addWidget(self, widget, size):
        self.addSpacing(size)
        return self.addWidgetWithoutSize(widget)

    def addLabel(self, text, justification='center'):
        label = Label(text,
                      self.nextLocation('small', justification),
                      justification)
        return self.addWidget(label, 'small')

    def addButton(self, text, onClick, closeMenu=True, justification='center'):
        if closeMenu:
            def fullOnClick():
                self.close()
                onClick()
        else:
            fullOnClick = onClick

        button = Button(text,
                        self.nextLocation('big', justification),
                        fullOnClick)
        return self.addWidget(button, 'big')

    def _addTagButton(self, location, tag, onClick, closeMenu=True,
            colorSchemeCallback=None,
            justification='center'):
        if closeMenu:
            def fullOnClick():
                self.close()
                onClick(tag)
        else:
            def fullOnClick():
                onClick(tag)

        if colorSchemeCallback:
            def csc():
                return colorSchemeCallback(tag)
        else:
            csc = None

        button = Button(str(tag), location, fullOnClick,
                        colorSchemeCallback=csc)
        return self.addWidgetWithoutSize(button)

    def addButtonRow(self, tags, onClick, closeMenu=True,
            colorSchemeCallback=None, size='big'):
        x,y = self.nextLocation(size)
        spacing = 50
        n = len(tags)
        for i in range(n):
            xx = x + spacing * (2 * i - (n-1))/2
            self._addTagButton((xx, y), tags[i], onClick, closeMenu,
                    colorSchemeCallback)
        self.addSpacing(size)

    def addButtonMultiRow(self, tags, onClick, closeMenu=True,
            colorSchemeCallback=None):
        rowSize = 10
        i = 0
        n = (len(tags)+rowSize-1)/rowSize
        for i in range(n):
            self.addButtonRow(tags[i*rowSize:(i+1)*rowSize], onClick,
                    closeMenu,
                    colorSchemeCallback = colorSchemeCallback,
                    size= 'big' if i == n-1 else 'small')

    def addReturnButton(self, text='Return', justification='center'):
        def onClick():
            # Closing the menu is done automatically.
            pass
        return self.addButton(text, onClick, closeMenu=True,
                              justification=justification)

    def addTextBox(self, capacity, text='', onEnter=None,
                   justification='center'):
        textBox = TextBox(self.nextLocation('big', justification),
                          capacity,
                          text,
                          onEnter)
        return self.addWidget(textBox, 'big')
        
# Init Menus {{{1
def colorSchemeForMap(filename):
    if filename in settings['beaten']:
        return beatenColorScheme
    elif filename in settings['played']:
        return playedColorScheme
    else:
        return standardColorScheme

@postInit
def initMenus():
    global mainMenu, resumeButton

    mainMenu = Menu()
    resumeButton = mainMenu.addButton('Resume Last Level', game.playCurrentMap)

    mainMenu.addSpacing('small')

    mainMenu.addLabel("Tutorials:")
    def tutorialCallback(i):
        game.playMap('tut'+i+'.map')
    def tutCSC(i):
        return colorSchemeForMap('tut'+i+'.map')
    mainMenu.addButtonMultiRow([str(i) for i in numbersByPrefix('tut')],
            tutorialCallback, colorSchemeCallback=tutCSC)

    mainMenu.addSpacing('small')

    mainMenu.addLabel("Levels:")
    def levelCallback(i):
        game.playMap('level'+i+'.map')
    def levelCSC(i):
        return colorSchemeForMap('level'+i+'.map')

    mainMenu.addButtonMultiRow([str(i) for i in numbersByPrefix('level')],
            levelCallback, colorSchemeCallback=levelCSC)
    def editorCallback():
        game.editCurrentMap()
    mainMenu.addButton('Editor', editorCallback)

    mainMenu.addButton('Exit', terminate)

    game.staticMap.suspend()
    showAppropriateWidgets()
    #mainMenu.show()

    global newbieMenu

    newbieMenu = Menu()

    def beginTutorial():
        game.playMap(allLevelFiles[0])
    newbieMenu.addSpacing('big')
    newbieMenu.addButton('Begin Tutorial', beginTutorial)
    newbieMenu.addSpacing('big')
    newbieMenu.addSpacing('big')
    newbieMenu.addButton('Main Menu', gotoMainMenu)

    global noplayMenu
    noplayMenu = Menu()

    noplayMenu.addLabel(
            "This game needs PyGame version 1.8 or greater to run correctly.")
    noplayMenu.addLabel(
            "Please uninstall PyGame version " + pygame.version.ver +
            " and install a more recent version.")
    noplayMenu.addSpacing('big')
    noplayMenu.addButton('Wisely Exit', terminate)
    def playAnyway():
        messageTicker.add("Don't say I didn't warn you.")
        # TODO: Remove duplicate code
        gotoMainMenu()
        if not settings['lastLevelPlayed']:
            newbieMenu.show()
    noplayMenu.addSpacing('big')
    noplayMenu.addButton('Foolishly Play Anyway', playAnyway)

    global victoryMenu
    victoryMenu = Menu()

    victoryMenu.addLabel("Congratulations!")
    victoryMenu.addSpacing('small')
    victoryMenu.addLabel("You have beaten all levels of the game.")
    victoryMenu.addLabel(
            "I hope you enjoyed playing it as much as I did creating it.")
    victoryMenu.addLabel(' '*50+"-Andy")
    victoryMenu.addSpacing('big')
    victoryMenu.addSpacing('big')
    victoryMenu.addButton('Main Menu', gotoMainMenu)

# Map autoloading {{{1
@postInit
def autoLoadMap():
    last = settings['lastLevelPlayed']
    if last:
        game.editMap(last)
    if pygame.version.vernum < (1,8) and platform.system() == 'Windows':
        noplayMenu.show()
    elif not last:
        newbieMenu.show()
#    if os.access('last.map', os.R_OK):
#        game.editMap('last.map')
#    game.playMap(levelFiles[0])

#@onExit
#def autoSaveMap():
#    map = game.staticMap
#    if map.__class__ == UserMap:
#        map = map.base
#    map.save('last.map')

# Display Code {{{1
def resetDisplay():
    video_flags = 0
    if settings['fullscreen']:
        video_flags |= FULLSCREEN

    global screen
    screen = pygame.display.set_mode((screenWidth,screenHeight), video_flags)

    pygame.display.set_caption("Powderkeg")
    icon = Sprite('gameicon')
    pygame.display.set_icon(icon.surface)

def toggleFullscreen():
    settings['fullscreen'] = not settings['fullscreen']
    resetDisplay()

def draw():
    #screen.fill((150,120,0))
    sprites.get('backdrop').draw((0,0))
    scene.draw()

# Event Handling Code {{{1
def onEscape():
    if currentMenu:
        return 0
    elif game.dynamicMap:
        replayLevel()
        return 1
    else:
        gotoMainMenu()
        return 1

def onKeyDown(event):
    global square
    if event.key == K_ESCAPE:
        return onEscape()
    elif clickables.key(event.unicode, event.key, event.mod):
        return 1
#    elif event.key == K_SPACE:
#        createEntity(PauseMessage, pause = not updater.paused)
#    elif event.key == K_q:
#        createEntity(QuitToMenuMessage)
    elif event.key == K_f:
        toggleFPS()
    elif event.key == K_m:
        mixer.toggleMusic()
    elif event.key == K_s:
        mixer.toggleSound()
#    elif event.key == K_d:
#        if len(towers) == 0:
#            startDemo()
#    elif event.key == K_z:
#        if len(towers) == 0:
#            startDarwinator()
    elif event.key == K_w:
        toggleFullscreen()

    return 1
    
def onMouseDown(event):
    if event.button == 1:
        clickables.click(event.pos)

def onMouseMotion(event):
    clickables.move(event.pos)

def onMouseUp(event):
    if event.button == 1:
        clickables.release(event.pos)

def onUserEvent(event):
    if event.type == PURGE_EVENT:
#        scene.purge()
        showFPS()
    else:
        assert event.type == MUSIC_EVENT
        mixer.newSong()

# Init {{{1
def init():
    pygame.mixer.pre_init(44100,-16,2, 1024 * 3)
    pygame.init()
    pygame.time.set_timer(PURGE_EVENT, 1000)

    resetDisplay()

    callPostInits()

    #showMainMenu()

# Main {{{1
def main():
    init()

    startTicks = pygame.time.get_ticks()
    lastTicks = startTicks

    playing = True

    while playing:
        ticks = pygame.time.get_ticks()
        clock.tick()

        #event = pygame.event.poll()

        for event in pygame.event.get():
            if event.type == QUIT:
                playing = False
            if event.type == KEYDOWN:
                if onKeyDown(event) == 0:
                    playing = False
            if event.type == MOUSEBUTTONDOWN:
                onMouseDown(event)
            if event.type == MOUSEMOTION:
                onMouseMotion(event)
            if event.type == MOUSEBUTTONUP:
                onMouseUp(event)
            if event.type >= USEREVENT:
                onUserEvent(event)

        duration = ticks - lastTicks
        if duration > updateThreshold:
            updater.update(ticks, duration)
            lastTicks = ticks

        pygame.event.pump()

        draw()
        pygame.display.flip()

    callOnExits()


if __name__ == '__main__': main()
